use crate::models::CreateTrackedPlaylist;
use crate::subsonic::Subsonic;
use crate::CONFIG;
use chrono::Local;
use sqlx::postgres::PgPool;
use std::str::FromStr;
use std::time::Duration;
use tokio::time::{interval, MissedTickBehavior};
use uuid::Uuid;

pub async fn create_playlists_daemon(
    media_client: impl AsRef<Subsonic> + Clone,
) -> Result<(), Box<dyn std::error::Error>> {
    println!("Create playlist daemon starting...");

    let database = PgPool::connect(
        CONFIG
            .get()
            .unwrap()
            .get::<String>("database.url")?
            .as_str(),
    )
    .await
    .unwrap();

    let mut ticker = interval(Duration::from_secs(
        CONFIG.get().unwrap().get::<u64>("daemon.interval")?,
    ));
    ticker.set_missed_tick_behavior(MissedTickBehavior::Skip);

    loop {
        let records_res = sqlx::query_as!(
            CreateTrackedPlaylist,
            r#"SELECT * FROM "create_tracked_playlist""#
        )
        .fetch_all(&database)
        .await;

        match records_res {
            Ok(records) => {
                for record in records {
                    // todo handle me
                    create_playlist(media_client.clone(), record, database.clone())
                        .await
                        .unwrap();
                }
            }

            Err(e) => {
                println!("Could not fetch create_tracked_playlist: {:?}", e);
            }
        }

        ticker.tick().await;
    }
}

async fn create_playlist(
    media_client: impl AsRef<Subsonic>,
    playlist_rule: CreateTrackedPlaylist,
    database: PgPool,
) -> Result<(), Box<dyn std::error::Error>> {
    let now = Local::now();
    let new_name = now.format(&playlist_rule.name_template).to_string();

    if Some(&new_name) != playlist_rule.last_name.as_ref() {
        // Create playlist
        let playlist = media_client.as_ref().create_playlist(&new_name).await?;
        let uuid = Uuid::new_v4();
        let playlist_uuid = Uuid::from_str(&playlist.id)?;

        sqlx::query!(
            r#"
            INSERT INTO "tracked_playlist" (id, playlist_id, playlist_size, tracking_user, tracking_type)
            VALUES ($1, $2, $3, $4, $5)
            "#,
            uuid,
            playlist_uuid,
            playlist_rule.playlist_size,
            playlist_rule.tracking_user,
            playlist_rule.tracking_type,
        )
        .execute(&database)
        .await?;

        sqlx::query!(
            r#"
            UPDATE "create_tracked_playlist" SET last_name = $1 WHERE id = $2
            "#,
            new_name,
            playlist_rule.id,
        )
        .execute(&database)
        .await?;
    }

    Ok(())
}
